/*
 * Copyright 2020 Tadashi G. Takaoka
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "table_mc68000.h"

#include "entry_mc68000.h"
#include "entry_table.h"
#include "text_mc68000.h"

using namespace libasm::text::mc68000;

namespace libasm {
namespace mc68000 {

#define E(_opc, _name, _isize, _src, _dst, _srcp, _dstp, _osize, _hasSize) \
    { _opc, Entry::Flags::create(_src, _dst, _srcp, _dstp, _osize, _isize, _hasSize), _name }
#define E2(_opc, _name, _isize, _src, _dst, _srcp, _dstp, _osize) \
    E(_opc, _name, _isize, _src, _dst, _srcp, _dstp, _osize, false)
#define E1(_opc, _name, _isize, _src, _srcp, _osize) \
    E2(_opc, _name, _isize, _src, M_NONE, _srcp, OP___, _osize)
#define E0(_opc, _name) E1(_opc, _name, ISZ_NONE, M_NONE, OP___, SZ_NONE)
#define W2(_opc, _name, _isize, _src, _dst, _srcp, _dstp, _osize) \
    E(_opc, _name, _isize, _src, _dst, _srcp, _dstp, _osize, true)
#define W1(_opc, _name, _isize, _src, _srcp, _osize) \
    W2(_opc, _name, _isize, _src, M_NONE, _srcp, OP___, _osize)
#define A2(_opc, _name, _isize, _src, _dst, _srcp, _dstp, _osize) \
    E(_opc, _name, _isize, _src, _dst, _srcp, _dstp, _osize, false)

// clang-format off
static constexpr Entry MC68000_TABLE[] PROGMEM = {
    E2(0000074, TEXT_ORI,   ISZ_NONE, M_IMDAT, M_CCR,   OP___, OP___, SZ_BYTE),
    E2(0000174, TEXT_ORI,   ISZ_NONE, M_IMDAT, M_SR,    OP___, OP___, SZ_WORD),
    E2(0000000, TEXT_ORI,   ISZ_DATA, M_IMDAT, M_WDATA, OP___, OP_10, SZ_DATA),
    E2(0001074, TEXT_ANDI,  ISZ_NONE, M_IMDAT, M_CCR,   OP___, OP___, SZ_BYTE),
    E2(0001174, TEXT_ANDI,  ISZ_NONE, M_IMDAT, M_SR,    OP___, OP___, SZ_WORD),
    E2(0001000, TEXT_ANDI,  ISZ_DATA, M_IMDAT, M_WDATA, OP___, OP_10, SZ_DATA),
    E2(0002000, TEXT_SUBI,  ISZ_DATA, M_IMDAT, M_WDATA, OP___, OP_10, SZ_DATA),
    E2(0003000, TEXT_ADDI,  ISZ_DATA, M_IMDAT, M_WDATA, OP___, OP_10, SZ_DATA),
    E2(0005074, TEXT_EORI,  ISZ_NONE, M_IMDAT, M_CCR,   OP___, OP___, SZ_BYTE),
    E2(0005174, TEXT_EORI,  ISZ_NONE, M_IMDAT, M_SR,    OP___, OP___, SZ_WORD),
    E2(0005000, TEXT_EORI,  ISZ_DATA, M_IMDAT, M_WDATA, OP___, OP_10, SZ_DATA),
    E2(0006000, TEXT_CMPI,  ISZ_DATA, M_IMDAT, M_WDATA, OP___, OP_10, SZ_DATA),
    W2(0004000, TEXT_BTST,  ISZ_NONE, M_IMBIT, M_DREG,  OP___, OP__0, SZ_QUAD),
    W2(0004000, TEXT_BTST,  ISZ_NONE, M_IMBIT, M_RMEM,  OP___, OP_10, SZ_BYTE),
    W2(0004100, TEXT_BCHG,  ISZ_NONE, M_IMBIT, M_DREG,  OP___, OP__0, SZ_QUAD),
    W2(0004100, TEXT_BCHG,  ISZ_NONE, M_IMBIT, M_WMEM,  OP___, OP_10, SZ_BYTE),
    W2(0004200, TEXT_BCLR,  ISZ_NONE, M_IMBIT, M_DREG,  OP___, OP__0, SZ_QUAD),
    W2(0004200, TEXT_BCLR,  ISZ_NONE, M_IMBIT, M_WMEM,  OP___, OP_10, SZ_BYTE),
    W2(0004300, TEXT_BSET,  ISZ_NONE, M_IMBIT, M_DREG,  OP___, OP__0, SZ_QUAD),
    W2(0004300, TEXT_BSET,  ISZ_NONE, M_IMBIT, M_WMEM,  OP___, OP_10, SZ_BYTE),
    E2(0000410, TEXT_MOVEP, ISZ_DATA, M_DISP,  M_DREG,  OP__0, OP__3, SZ_ADDR),
    E2(0000610, TEXT_MOVEP, ISZ_DATA, M_DREG,  M_DISP,  OP__3, OP__0, SZ_ADDR),
    W2(0000400, TEXT_BTST,  ISZ_NONE, M_DREG,  M_DREG,  OP__3, OP__0, SZ_QUAD),
    W2(0000400, TEXT_BTST,  ISZ_NONE, M_DREG,  M_RMEM,  OP__3, OP_10, SZ_BYTE),
    W2(0000500, TEXT_BCHG,  ISZ_NONE, M_DREG,  M_DREG,  OP__3, OP__0, SZ_QUAD),
    W2(0000500, TEXT_BCHG,  ISZ_NONE, M_DREG,  M_WMEM,  OP__3, OP_10, SZ_BYTE),
    W2(0000600, TEXT_BCLR,  ISZ_NONE, M_DREG,  M_DREG,  OP__3, OP__0, SZ_QUAD),
    W2(0000600, TEXT_BCLR,  ISZ_NONE, M_DREG,  M_WMEM,  OP__3, OP_10, SZ_BYTE),
    W2(0000700, TEXT_BSET,  ISZ_NONE, M_DREG,  M_DREG,  OP__3, OP__0, SZ_QUAD),
    W2(0000700, TEXT_BSET,  ISZ_NONE, M_DREG,  M_WMEM,  OP__3, OP_10, SZ_BYTE),
    E2(0020100, TEXT_MOVEA, ISZ_QUAD, M_RADDR, M_AREG,  OP_10, OP__3, SZ_QUAD),
    E2(0030100, TEXT_MOVEA, ISZ_WORD, M_RADDR, M_AREG,  OP_10, OP__3, SZ_WORD),
    E2(0030100, TEXT_MOVEA, ISZ_NONE, M_RADDR, M_AREG,  OP_10, OP__3, SZ_WORD),
    W2(0042300, TEXT_MOVE,  ISZ_NONE, M_RDATA, M_CCR,   OP_10, OP___, SZ_WORD),
    W2(0040300, TEXT_MOVE,  ISZ_NONE, M_SR,    M_WDATA, OP___, OP_10, SZ_WORD),
    W2(0043300, TEXT_MOVE,  ISZ_NONE, M_RDATA, M_SR,    OP_10, OP___, SZ_WORD),
    E2(0010000, TEXT_MOVE,  ISZ_BYTE, M_RDATA, M_WDATA, OP_10, OP_23, SZ_BYTE),
    E2(0020000, TEXT_MOVE,  ISZ_QUAD, M_RADDR, M_WDATA, OP_10, OP_23, SZ_QUAD),
    E2(0030000, TEXT_MOVE,  ISZ_WORD, M_RADDR, M_WDATA, OP_10, OP_23, SZ_WORD),
    E1(0040000, TEXT_NEGX,  ISZ_DATA, M_WDATA, OP_10, SZ_DATA),
    E1(0041000, TEXT_CLR,   ISZ_DATA, M_WDATA, OP_10, SZ_DATA),
    E1(0042000, TEXT_NEG,   ISZ_DATA, M_WDATA, OP_10, SZ_DATA),
    E1(0043000, TEXT_NOT,   ISZ_DATA, M_WDATA, OP_10, SZ_DATA),
    E1(0044200, TEXT_EXT,   ISZ_DATA, M_DREG,  OP__0, SZ_ADDR),
    E1(0044000, TEXT_NBCD,  ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0044100, TEXT_SWAP,  ISZ_NONE, M_DREG,  OP__0, SZ_QUAD),
    E1(0044100, TEXT_PEA,   ISZ_NONE, M_JADDR, OP_10, SZ_QUAD),
    E0(0045374, TEXT_ILLEG),
    E1(0045300, TEXT_TAS,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0045000, TEXT_TST,   ISZ_DATA, M_WDATA, OP_10, SZ_DATA),
    E1(0047100, TEXT_TRAP,  ISZ_NONE, M_IMVEC, OP___, SZ_NONE),
    E2(0047120, TEXT_LINK,  ISZ_NONE, M_AREG,  M_IMDSP, OP__0, OP___, SZ_WORD),
    E1(0047130, TEXT_UNLK,  ISZ_NONE, M_AREG,  OP__0, SZ_NONE),
    E2(0047140, TEXT_MOVE,  ISZ_NONE, M_AREG,  M_USP,   OP__0, OP___, SZ_QUAD),
    E2(0047150, TEXT_MOVE,  ISZ_NONE, M_USP,   M_AREG,  OP___, OP__0, SZ_QUAD),
    E0(0047160, TEXT_RESET),
    E0(0047161, TEXT_NOP),
    E1(0047162, TEXT_STOP,  ISZ_NONE, M_IMDAT, OP___, SZ_WORD),
    E0(0047163, TEXT_RTE),
    E0(0047165, TEXT_RTS),
    E0(0047166, TEXT_TRAPV),
    E0(0047167, TEXT_RTR),
    E1(0047200, TEXT_JSR,   ISZ_NONE, M_JADDR, OP_10, SZ_WORD),
    E1(0047300, TEXT_JMP,   ISZ_NONE, M_JADDR, OP_10, SZ_WORD),
    E2(0044200, TEXT_MOVEM, ISZ_DATA, M_MULT,  M_DADDR, OP___, OP_10, SZ_ADDR),
    E2(0046200, TEXT_MOVEM, ISZ_DATA, M_IADDR, M_MULT,  OP_10, OP___, SZ_ADDR),
    E2(0040700, TEXT_LEA,   ISZ_NONE, M_JADDR, M_AREG,  OP_10, OP__3, SZ_QUAD),
    E2(0040600, TEXT_CHK,   ISZ_WORD, M_RDATA, M_DREG,  OP_10, OP__3, SZ_WORD),
    E2(0050310, TEXT_DBT,   ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0050710, TEXT_DBRA,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0050710, TEXT_DBF,   ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0051310, TEXT_DBHI,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0051710, TEXT_DBLS,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0052310, TEXT_DBCC,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0052310, TEXT_DBHS,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0052710, TEXT_DBCS,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0052710, TEXT_DBLO,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0053310, TEXT_DBNE,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0053710, TEXT_DBEQ,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0054310, TEXT_DBVC,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0054710, TEXT_DBVS,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0055310, TEXT_DBPL,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0055710, TEXT_DBMI,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0056310, TEXT_DBGE,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0056710, TEXT_DBLT,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0057310, TEXT_DBGT,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E2(0057710, TEXT_DBLE,  ISZ_NONE, M_DREG,  M_REL16, OP__0, OP___, SZ_WORD),
    E1(0050300, TEXT_ST,    ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0050700, TEXT_SF,    ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0051300, TEXT_SHI,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0051700, TEXT_SLS,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0052300, TEXT_SCC,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0052300, TEXT_SHS,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0052700, TEXT_SCS,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0052700, TEXT_SLO,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0053300, TEXT_SNE,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0053700, TEXT_SEQ,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0054300, TEXT_SVC,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0054700, TEXT_SVS,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0055300, TEXT_SPL,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0055700, TEXT_SMI,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0056300, TEXT_SGE,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0056700, TEXT_SLT,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0057300, TEXT_SGT,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E1(0057700, TEXT_SLE,   ISZ_NONE, M_WDATA, OP_10, SZ_BYTE),
    E2(0050000, TEXT_ADDQ,  ISZ_DATA, M_IM3,   M_WADDR, OP___, OP_10, SZ_DATA),
    E2(0050400, TEXT_SUBQ,  ISZ_DATA, M_IM3,   M_WADDR, OP___, OP_10, SZ_DATA),
    E1(0060000, TEXT_BRA,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0060000, TEXT_BRA,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0060000, TEXT_BT,    ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0060000, TEXT_BT,    ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0060400, TEXT_BSR,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0060400, TEXT_BSR,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0061000, TEXT_BHI,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0061000, TEXT_BHI,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0061400, TEXT_BLS,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0061400, TEXT_BLS,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0062000, TEXT_BCC,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0062000, TEXT_BCC,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0062000, TEXT_BHS,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0062000, TEXT_BHS,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0062400, TEXT_BCS,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0062400, TEXT_BCS,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0062400, TEXT_BLO,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0062400, TEXT_BLO,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0063000, TEXT_BNE,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0063000, TEXT_BNE,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0063400, TEXT_BEQ,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0063400, TEXT_BEQ,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0064000, TEXT_BVC,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0064000, TEXT_BVC,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0064400, TEXT_BVS,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0064400, TEXT_BVS,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0065000, TEXT_BPL,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0065000, TEXT_BPL,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0065400, TEXT_BMI,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0065400, TEXT_BMI,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0066000, TEXT_BGE,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0066000, TEXT_BGE,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0066400, TEXT_BLT,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0066400, TEXT_BLT,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0067000, TEXT_BGT,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0067000, TEXT_BGT,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E1(0067400, TEXT_BLE,   ISZ_NONE, M_REL8,  OP___, SZ_BYTE),
    E1(0067400, TEXT_BLE,   ISZ_NONE, M_REL16, OP___, SZ_WORD),
    E2(0070000, TEXT_MOVEQ, ISZ_NONE, M_IM8,   M_DREG,  OP___, OP__3, SZ_QUAD),
    E2(0100300, TEXT_DIVU,  ISZ_DATA, M_RDATA, M_DREG,  OP_10, OP__3, SZ_WORD),
    E2(0100700, TEXT_DIVS,  ISZ_DATA, M_RDATA, M_DREG,  OP_10, OP__3, SZ_WORD),
    E2(0100000, TEXT_OR,    ISZ_DATA, M_RDATA, M_DREG,  OP_10, OP__3, SZ_DATA),
    E2(0100400, TEXT_SBCD,  ISZ_NONE, M_DREG,  M_DREG,  OP__0, OP__3, SZ_BYTE),
    E2(0100410, TEXT_SBCD,  ISZ_NONE, M_PDEC,  M_PDEC,  OP__0, OP__3, SZ_BYTE),
    E2(0100400, TEXT_OR,    ISZ_DATA, M_DREG,  M_WMEM,  OP__3, OP_10, SZ_DATA),
    E2(0110300, TEXT_SUBA,  ISZ_DATA, M_RADDR, M_AREG,  OP_10, OP__3, SZ_ADR8),
    E2(0110400, TEXT_SUBX,  ISZ_DATA, M_DREG,  M_DREG,  OP__0, OP__3, SZ_DATA),
    E2(0110410, TEXT_SUBX,  ISZ_DATA, M_PDEC,  M_PDEC,  OP__0, OP__3, SZ_DATA),
    E2(0110000, TEXT_SUB,   ISZ_DATA, M_RADDR, M_DREG,  OP_10, OP__3, SZ_DATA),
    E2(0110400, TEXT_SUB,   ISZ_DATA, M_DREG,  M_WMEM,  OP__3, OP_10, SZ_DATA),
    E2(0130300, TEXT_CMPA,  ISZ_DATA, M_RADDR, M_AREG,  OP_10, OP__3, SZ_ADR8),
    E2(0130000, TEXT_CMP,   ISZ_DATA, M_RADDR, M_DREG,  OP_10, OP__3, SZ_DATA),
    E2(0130410, TEXT_CMPM,  ISZ_DATA, M_PINC,  M_PINC,  OP__0, OP__3, SZ_DATA),
    E2(0130400, TEXT_EOR,   ISZ_DATA, M_DREG,  M_WDATA, OP__3, OP_10, SZ_DATA),
    E2(0140300, TEXT_MULU,  ISZ_DATA, M_RDATA, M_DREG,  OP_10, OP__3, SZ_WORD),
    E2(0140700, TEXT_MULS,  ISZ_DATA, M_RDATA, M_DREG,  OP_10, OP__3, SZ_WORD),
    E2(0140000, TEXT_AND,   ISZ_DATA, M_RDATA, M_DREG,  OP_10, OP__3, SZ_DATA),
    E2(0140400, TEXT_ABCD,  ISZ_NONE, M_DREG,  M_DREG,  OP__0, OP__3, SZ_BYTE),
    E2(0140410, TEXT_ABCD,  ISZ_NONE, M_PDEC,  M_PDEC,  OP__0, OP__3, SZ_BYTE),
    E2(0140500, TEXT_EXG,   ISZ_NONE, M_DREG,  M_DREG,  OP__3, OP__0, SZ_QUAD),
    E2(0140510, TEXT_EXG,   ISZ_NONE, M_AREG,  M_AREG,  OP__3, OP__0, SZ_QUAD),
    E2(0140610, TEXT_EXG,   ISZ_NONE, M_DREG,  M_AREG,  OP__3, OP__0, SZ_QUAD),
    E2(0140610, TEXT_EXG,   ISZ_NONE, M_AREG,  M_DREG,  OP__0, OP__3, SZ_QUAD),
    E2(0140400, TEXT_AND,   ISZ_DATA, M_DREG,  M_WMEM,  OP__3, OP_10, SZ_DATA),
    E2(0150300, TEXT_ADDA,  ISZ_DATA, M_RADDR, M_AREG,  OP_10, OP__3, SZ_ADR8),
    E2(0150400, TEXT_ADDX,  ISZ_DATA, M_DREG,  M_DREG,  OP__0, OP__3, SZ_DATA),
    E2(0150410, TEXT_ADDX,  ISZ_DATA, M_PDEC,  M_PDEC,  OP__0, OP__3, SZ_DATA),
    E2(0150000, TEXT_ADD,   ISZ_DATA, M_RADDR, M_DREG,  OP_10, OP__3, SZ_DATA),
    E2(0150400, TEXT_ADD,   ISZ_DATA, M_DREG,  M_WMEM,  OP__3, OP_10, SZ_DATA),
    W1(0160300, TEXT_ASR,   ISZ_NONE, M_WMEM,  OP_10, SZ_WORD),
    W1(0160700, TEXT_ASL,   ISZ_NONE, M_WMEM,  OP_10, SZ_WORD),
    W1(0161300, TEXT_LSR,   ISZ_NONE, M_WMEM,  OP_10, SZ_WORD),
    W1(0161700, TEXT_LSL,   ISZ_NONE, M_WMEM,  OP_10, SZ_WORD),
    W1(0162300, TEXT_ROXR,  ISZ_NONE, M_WMEM,  OP_10, SZ_WORD),
    W1(0162700, TEXT_ROXL,  ISZ_NONE, M_WMEM,  OP_10, SZ_WORD),
    W1(0163300, TEXT_ROR,   ISZ_NONE, M_WMEM,  OP_10, SZ_WORD),
    W1(0163700, TEXT_ROL,   ISZ_NONE, M_WMEM,  OP_10, SZ_WORD),
    E2(0160000, TEXT_ASR,   ISZ_DATA, M_IM3,   M_DREG,  OP___, OP__0, SZ_DATA),
    E2(0160040, TEXT_ASR,   ISZ_DATA, M_DREG,  M_DREG,  OP__3, OP__0, SZ_DATA),
    E2(0160400, TEXT_ASL,   ISZ_DATA, M_IM3,   M_DREG,  OP___, OP__0, SZ_DATA),
    E2(0160440, TEXT_ASL,   ISZ_DATA, M_DREG,  M_DREG,  OP__3, OP__0, SZ_DATA),
    E2(0160010, TEXT_LSR,   ISZ_DATA, M_IM3,   M_DREG,  OP___, OP__0, SZ_DATA),
    E2(0160050, TEXT_LSR,   ISZ_DATA, M_DREG,  M_DREG,  OP__3, OP__0, SZ_DATA),
    E2(0160410, TEXT_LSL,   ISZ_DATA, M_IM3,   M_DREG,  OP___, OP__0, SZ_DATA),
    E2(0160450, TEXT_LSL,   ISZ_DATA, M_DREG,  M_DREG,  OP__3, OP__0, SZ_DATA),
    E2(0160020, TEXT_ROXR,  ISZ_DATA, M_IM3,   M_DREG,  OP___, OP__0, SZ_DATA),
    E2(0160060, TEXT_ROXR,  ISZ_DATA, M_DREG,  M_DREG,  OP__3, OP__0, SZ_DATA),
    E2(0160420, TEXT_ROXL,  ISZ_DATA, M_IM3,   M_DREG,  OP___, OP__0, SZ_DATA),
    E2(0160460, TEXT_ROXL,  ISZ_DATA, M_DREG,  M_DREG,  OP__3, OP__0, SZ_DATA),
    E2(0160030, TEXT_ROR,   ISZ_DATA, M_IM3,   M_DREG,  OP___, OP__0, SZ_DATA),
    E2(0160070, TEXT_ROR,   ISZ_DATA, M_DREG,  M_DREG,  OP__3, OP__0, SZ_DATA),
    E2(0160430, TEXT_ROL,   ISZ_DATA, M_IM3,   M_DREG,  OP___, OP__0, SZ_DATA),
    E2(0160470, TEXT_ROL,   ISZ_DATA, M_DREG,  M_DREG,  OP__3, OP__0, SZ_DATA),
    A2(0020100, TEXT_MOVE,  ISZ_QUAD, M_RADDR, M_AREG,  OP_10, OP__3, SZ_QUAD),
    A2(0030100, TEXT_MOVE,  ISZ_WORD, M_RADDR, M_AREG,  OP_10, OP__3, SZ_WORD),
    A2(0110300, TEXT_SUB,   ISZ_DATA, M_RADDR, M_AREG,  OP_10, OP__3, SZ_ADR8),
    A2(0110374, TEXT_SUBI,  ISZ_DATA, M_IMDAT, M_AREG,  OP___, OP__3, SZ_ADR8),
    A2(0130374, TEXT_CMPI,  ISZ_DATA, M_IMDAT, M_AREG,  OP___, OP__3, SZ_ADR8),
    A2(0150300, TEXT_ADD,   ISZ_DATA, M_RADDR, M_AREG,  OP_10, OP__3, SZ_ADR8),
    A2(0150374, TEXT_ADDI,  ISZ_DATA, M_IMDAT, M_AREG,  OP___, OP__3, SZ_ADR8),
};

static constexpr uint8_t MC68000_INDEX[] PROGMEM = {
    164,  // TEXT_ABCD
    165,  // TEXT_ABCD
    174,  // TEXT_ADD
    175,  // TEXT_ADD
    205,  // TEXT_ADD
    171,  // TEXT_ADDA
      7,  // TEXT_ADDI
    206,  // TEXT_ADDI
    105,  // TEXT_ADDQ
    172,  // TEXT_ADDX
    173,  // TEXT_ADDX
    163,  // TEXT_AND
    170,  // TEXT_AND
      3,  // TEXT_ANDI
      4,  // TEXT_ANDI
      5,  // TEXT_ANDI
    177,  // TEXT_ASL
    186,  // TEXT_ASL
    187,  // TEXT_ASL
    176,  // TEXT_ASR
    184,  // TEXT_ASR
    185,  // TEXT_ASR
    117,  // TEXT_BCC
    118,  // TEXT_BCC
     14,  // TEXT_BCHG
     15,  // TEXT_BCHG
     24,  // TEXT_BCHG
     25,  // TEXT_BCHG
     16,  // TEXT_BCLR
     17,  // TEXT_BCLR
     26,  // TEXT_BCLR
     27,  // TEXT_BCLR
    121,  // TEXT_BCS
    122,  // TEXT_BCS
    127,  // TEXT_BEQ
    128,  // TEXT_BEQ
    137,  // TEXT_BGE
    138,  // TEXT_BGE
    141,  // TEXT_BGT
    142,  // TEXT_BGT
    113,  // TEXT_BHI
    114,  // TEXT_BHI
    119,  // TEXT_BHS
    120,  // TEXT_BHS
    143,  // TEXT_BLE
    144,  // TEXT_BLE
    123,  // TEXT_BLO
    124,  // TEXT_BLO
    115,  // TEXT_BLS
    116,  // TEXT_BLS
    139,  // TEXT_BLT
    140,  // TEXT_BLT
    135,  // TEXT_BMI
    136,  // TEXT_BMI
    125,  // TEXT_BNE
    126,  // TEXT_BNE
    133,  // TEXT_BPL
    134,  // TEXT_BPL
    107,  // TEXT_BRA
    108,  // TEXT_BRA
     18,  // TEXT_BSET
     19,  // TEXT_BSET
     28,  // TEXT_BSET
     29,  // TEXT_BSET
    111,  // TEXT_BSR
    112,  // TEXT_BSR
    109,  // TEXT_BT
    110,  // TEXT_BT
     12,  // TEXT_BTST
     13,  // TEXT_BTST
     22,  // TEXT_BTST
     23,  // TEXT_BTST
    129,  // TEXT_BVC
    130,  // TEXT_BVC
    131,  // TEXT_BVS
    132,  // TEXT_BVS
     67,  // TEXT_CHK
     40,  // TEXT_CLR
    158,  // TEXT_CMP
    157,  // TEXT_CMPA
     11,  // TEXT_CMPI
    204,  // TEXT_CMPI
    159,  // TEXT_CMPM
     73,  // TEXT_DBCC
     75,  // TEXT_DBCS
     78,  // TEXT_DBEQ
     70,  // TEXT_DBF
     83,  // TEXT_DBGE
     85,  // TEXT_DBGT
     71,  // TEXT_DBHI
     74,  // TEXT_DBHS
     86,  // TEXT_DBLE
     76,  // TEXT_DBLO
     72,  // TEXT_DBLS
     84,  // TEXT_DBLT
     82,  // TEXT_DBMI
     77,  // TEXT_DBNE
     81,  // TEXT_DBPL
     69,  // TEXT_DBRA
     68,  // TEXT_DBT
     79,  // TEXT_DBVC
     80,  // TEXT_DBVS
    147,  // TEXT_DIVS
    146,  // TEXT_DIVU
    160,  // TEXT_EOR
      8,  // TEXT_EORI
      9,  // TEXT_EORI
     10,  // TEXT_EORI
    166,  // TEXT_EXG
    167,  // TEXT_EXG
    168,  // TEXT_EXG
    169,  // TEXT_EXG
     43,  // TEXT_EXT
     47,  // TEXT_ILLEG
     63,  // TEXT_JMP
     62,  // TEXT_JSR
     66,  // TEXT_LEA
     51,  // TEXT_LINK
    179,  // TEXT_LSL
    190,  // TEXT_LSL
    191,  // TEXT_LSL
    178,  // TEXT_LSR
    188,  // TEXT_LSR
    189,  // TEXT_LSR
     33,  // TEXT_MOVE
     34,  // TEXT_MOVE
     35,  // TEXT_MOVE
     36,  // TEXT_MOVE
     37,  // TEXT_MOVE
     38,  // TEXT_MOVE
     53,  // TEXT_MOVE
     54,  // TEXT_MOVE
    200,  // TEXT_MOVE
    201,  // TEXT_MOVE
     30,  // TEXT_MOVEA
     31,  // TEXT_MOVEA
     32,  // TEXT_MOVEA
     64,  // TEXT_MOVEM
     65,  // TEXT_MOVEM
     20,  // TEXT_MOVEP
     21,  // TEXT_MOVEP
    145,  // TEXT_MOVEQ
    162,  // TEXT_MULS
    161,  // TEXT_MULU
     44,  // TEXT_NBCD
     41,  // TEXT_NEG
     39,  // TEXT_NEGX
     56,  // TEXT_NOP
     42,  // TEXT_NOT
    148,  // TEXT_OR
    151,  // TEXT_OR
      0,  // TEXT_ORI
      1,  // TEXT_ORI
      2,  // TEXT_ORI
     46,  // TEXT_PEA
     55,  // TEXT_RESET
    183,  // TEXT_ROL
    198,  // TEXT_ROL
    199,  // TEXT_ROL
    182,  // TEXT_ROR
    196,  // TEXT_ROR
    197,  // TEXT_ROR
    181,  // TEXT_ROXL
    194,  // TEXT_ROXL
    195,  // TEXT_ROXL
    180,  // TEXT_ROXR
    192,  // TEXT_ROXR
    193,  // TEXT_ROXR
     58,  // TEXT_RTE
     61,  // TEXT_RTR
     59,  // TEXT_RTS
    149,  // TEXT_SBCD
    150,  // TEXT_SBCD
     91,  // TEXT_SCC
     93,  // TEXT_SCS
     96,  // TEXT_SEQ
     88,  // TEXT_SF
    101,  // TEXT_SGE
    103,  // TEXT_SGT
     89,  // TEXT_SHI
     92,  // TEXT_SHS
    104,  // TEXT_SLE
     94,  // TEXT_SLO
     90,  // TEXT_SLS
    102,  // TEXT_SLT
    100,  // TEXT_SMI
     95,  // TEXT_SNE
     99,  // TEXT_SPL
     87,  // TEXT_ST
     57,  // TEXT_STOP
    155,  // TEXT_SUB
    156,  // TEXT_SUB
    202,  // TEXT_SUB
    152,  // TEXT_SUBA
      6,  // TEXT_SUBI
    203,  // TEXT_SUBI
    106,  // TEXT_SUBQ
    153,  // TEXT_SUBX
    154,  // TEXT_SUBX
     97,  // TEXT_SVC
     98,  // TEXT_SVS
     45,  // TEXT_SWAP
     48,  // TEXT_TAS
     50,  // TEXT_TRAP
     60,  // TEXT_TRAPV
     49,  // TEXT_TST
     52,  // TEXT_UNLK
};
// clang-format on

using EntryPage = entry::TableBase<Entry>;

using Cpu = entry::CpuBase<CpuType, EntryPage>;

static constexpr EntryPage MC68000_PAGES[] PROGMEM = {
        {ARRAY_RANGE(MC68000_TABLE), ARRAY_RANGE(MC68000_INDEX)},
};

static constexpr Cpu CPU_TABLE[] PROGMEM = {
        {MC68000, TEXT_CPU_68000, ARRAY_RANGE(MC68000_PAGES)},
};

static const Cpu *cpu(CpuType cpuType) {
    UNUSED(cpuType);
    return &CPU_TABLE[0];
}

static bool acceptMode(AddrMode opr, AddrMode table) {
    if (opr == table)
        return true;
    if (opr == M_DREG)
        return table == M_RADDR || table == M_RDATA || table == M_WADDR || table == M_WDATA ||
               table == M_MULT;
    if (opr == M_AREG)
        return table == M_RADDR || table == M_WADDR || table == M_MULT;
    if (opr == M_AIND || opr == M_DISP || opr == M_INDX || opr == M_AWORD || opr == M_ALONG)
        return table == M_RADDR || table == M_RDATA || table == M_WADDR || table == M_WDATA ||
               table == M_RMEM || table == M_WMEM || table == M_JADDR || table == M_IADDR ||
               table == M_DADDR;
    if (opr == M_PINC)
        return table == M_RADDR || table == M_RDATA || table == M_WADDR || table == M_WDATA ||
               table == M_RMEM || table == M_WMEM || table == M_IADDR;
    if (opr == M_PDEC)
        return table == M_RADDR || table == M_RDATA || table == M_WADDR || table == M_WDATA ||
               table == M_RMEM || table == M_WMEM || table == M_DADDR;
    if (opr == M_PCDSP || opr == M_PCIDX)
        return table == M_RADDR || table == M_RDATA || table == M_RMEM || table == M_JADDR ||
               table == M_IADDR;
    if (opr == M_IMDAT)
        return table == M_RADDR || table == M_RDATA || table == M_IMBIT || table == M_IM3 ||
               table == M_IM8 || table == M_IMVEC || table == M_IMDSP;
    if (opr == M_LABEL)
        return table == M_REL16 || table == M_REL8;
    return false;
}

static bool acceptSize(InsnSize insn, OprSize table, InsnSize isize) {
    const OprSize opr = OprSize(insn);
    if (opr == table)
        return true;
    if (opr == SZ_BYTE)
        return table == SZ_DATA;
    if (opr == SZ_WORD)
        return table == SZ_DATA || table == SZ_ADDR || table == SZ_ADR8;
    if (opr == SZ_QUAD)
        return table == SZ_DATA || table == SZ_ADDR || table == SZ_ADR8;
    if (opr == SZ_NONE)
        return table == SZ_WORD || table == SZ_DATA || isize == ISZ_NONE || table == SZ_ADDR ||
               table == SZ_ADR8;
    return false;
}

static bool acceptModes(AsmInsn &insn, const Entry *entry) {
    const auto table = entry->flags();
    return acceptMode(insn.srcOp.mode, table.src()) && acceptMode(insn.dstOp.mode, table.dst()) &&
           acceptSize(insn.insnSize(), table.oprSize(), table.insnSize());
}

Error TableMc68000::searchName(CpuType cpuType, AsmInsn &insn) const {
    cpu(cpuType)->searchName(insn, acceptModes);
    return insn.getError();
}

static Config::opcode_t getInsnMask(AddrMode src) {
    if (src == M_IM8 || src == M_REL8)
        return 0xFF;
    if (src == M_IMVEC)
        return 0xF;
    if (src == M_IM3)
        return 07000;
    return 0;
}

static Config::opcode_t getInsnMask(OprPos pos) {
    switch (pos) {
    case OP_10:
        return 00077;
    case OP_23:
        return 07700;
    case OP__0:
        return 00007;
    case OP__3:
        return 07000;
    default:
        return 0;
    }
}

static Config::opcode_t getInsnMask(OprSize size) {
    switch (size) {
    case SZ_DATA:
        return (3 << 6);
    case SZ_ADDR:
        return (1 << 6);
    case SZ_ADR8:
        return (1 << 8);
    default:
        return 0;
    }
}

static Config::opcode_t getInsnMask(Entry::Flags flags) {
    return getInsnMask(flags.src()) | getInsnMask(flags.srcPos()) | getInsnMask(flags.dstPos()) |
           getInsnMask(flags.oprSize());
}

static bool invalidModeReg(Config::opcode_t opc, AddrMode addrMode, OprPos pos, OprSize size) {
    uint8_t mode, reg;
    if (pos == OP_10) {
        mode = (opc >> 3) & 7;
        reg = (opc >> 0) & 7;
    } else if (pos == OP_23) {
        mode = (opc >> 6) & 7;
        reg = (opc >> 9) & 7;
    } else {
        return false;
    }
    if (mode < 7) {
        // mode
        // |  0 |  1 |   2  |   3   |   4   |    5   |     6     |
        // +----+----+------+-------+-------+--------+-----------+
        // | Dn |  1 | (An) | (An)+ | -(An) | (n,An) | (n,An,Xn) |
        if (addrMode == M_RDATA || addrMode == M_WDATA)
            return mode == 1;
        // |  0 |  1 | (An) | (An)+ | -(An) | (*,An) | (*,An,Xn) |
        if (addrMode == M_RMEM || addrMode == M_WMEM)
            return mode < 2;
        // |  0 |  1 | (An) |   3   |   4   | (n,An) | (n,An,Xn) |
        if (addrMode == M_JADDR)
            return mode < 2 || mode == 3 || mode == 4;
        // |  0 |  1 | (An) | (An)+ |   4   | (n,An) | (n,An,Xn) |
        if (addrMode == M_IADDR)
            return mode < 2 || mode == 4;
        //       (An)/     /-(An)/(*,An)/(*,An,Xn)
        // |  0 |  1 | (An) |   3   | -(An) | (n,An) | (n,An,Xn) |
        if (addrMode == M_DADDR)
            return mode < 2 || mode == 3;
        // | Dn | An*| (An) | (An)+ | -(An) | (n,An) | (n,An,Xn) |
        // M_WADDR, M_RADDR
        if (mode == 1 && size == SZ_DATA) {  // no BYTE size for An
            const auto insnSize = (opc >> 6) & 3;
            return insnSize == 0 || insnSize == 3;
        }
        return false;
    }
    // mode == 7
    if (reg >= 5)
        return true;
    // reg
    // |   0   |   1   |    2   |     3     |  4 |
    // | (n).W | (n).L | (n,PC) | (n,PC,Xn) | #n |
    if (addrMode == M_RADDR || addrMode == M_RDATA)
        return false;
    // | (n).W | (n).L | (n,PC) | (n,PC,Xn) |  4 |
    if (addrMode == M_RMEM || addrMode == M_JADDR || addrMode == M_IADDR)
        return reg == 4;
    // | (n).W | (n).L |    2   |     3     |  4 |
    // M_WADDR, M_WDATA, M_WMEM, M_DADDR
    return reg >= 2;
    return false;
}

static bool invalidSize(Config::opcode_t opc, OprSize size) {
    return size == SZ_DATA && ((opc >> 6) & 3) == 3;
}

static bool matchOpCode(DisInsn &insn, const Entry *entry, const EntryPage *page) {
    UNUSED(page);
    auto opc = insn.opCode();
    const auto flags = entry->flags();
    if (invalidModeReg(opc, flags.src(), flags.srcPos(), flags.oprSize()))
        return false;
    if (invalidModeReg(opc, flags.dst(), flags.dstPos(), flags.oprSize()))
        return false;
    if (invalidSize(opc, flags.oprSize()))
        return false;
    opc &= ~getInsnMask(flags);
    return opc == entry->opCode();
}

Error TableMc68000::searchOpCode(CpuType cpuType, DisInsn &insn, StrBuffer &out) const {
    cpu(cpuType)->searchOpCode(insn, out, matchOpCode);
    if (insn.getError() == UNKNOWN_INSTRUCTION)
        insn.nameBuffer().reset();
    return insn.getError();
}

const /*PROGMEM*/ char *TableMc68000::listCpu_P() const {
    return TEXT_CPU_MC68000;
}

const /*PROGMEM*/ char *TableMc68000::cpuName_P(CpuType cpuType) const {
    return cpu(cpuType)->name_P();
}

Error TableMc68000::searchCpuName(StrScanner &name, CpuType &cpuType) const {
    name.iexpectText_P(TEXT_CPU_MC68000, 2);
    if (name.iequals_P(TEXT_CPU_68000)) {
        cpuType = MC68000;
        return OK;
    }
    return UNSUPPORTED_CPU;
}

const TableMc68000 TABLE;

}  // namespace mc68000
}  // namespace libasm

// Local Variables:
// mode: c++
// c-basic-offset: 4
// tab-width: 4
// End:
// vim: set ft=cpp et ts=4 sw=4:
